if not BGControlPublic then BGControlPublic = {} end
if not BG_Control_SV then BG_Control_SV = {} end
if not BG_Control_SV.Options then BG_Control_SV.Options = {} end

if not BG_Control_SV.TalentCache then BG_Control_SV.TalentCache = {} end
if not BG_Control_SV.MMRCache then BG_Control_SV.MMRCache = {} end
if not BG_Control_SV.NodeCache then BG_Control_SV.NodeCache = {} end


-- Local hidden frames, functions, and everything else
-- TODO: ADD LOCAL HERE
BGC = {
	frame = CreateFrame("Frame","BGC_Frame"),
	MAX_RADIUS = 120,
	ADDON_PREFIX = "BG_CONTROL",
	ADDON_MSG_SEPERATOR = ","
	};

-- The "live" current BG Player Database for where people are located and who did what
-- TODO: Add Local HERE
PD = nil

function BGC.Player_Entering_Battleground(...)
	local _, type, _, _, _, _, _, mapID = GetInstanceInfo()
	if "pvp" == type then
		BGC.DebugPrint("Player has entered PVP Battleground ID:".. BGC.BGMapID())
		BGC.frame:RegisterEvent("CHAT_MSG_ADDON");
		BGC.frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
		BGC.frame:RegisterEvent("PLAYER_LEAVING_WORLD");
		BGC.frame:RegisterEvent("GROUP_ROSTER_UPDATE");	
		BGC.frame:RegisterEvent("UPDATE_BATTLEFIELD_SCORE");
		BGC.frame:RegisterEvent("WORLD_MAP_UPDATE");
	end
		
end

-- Well, if we WERE in PVP... and we leave... we're out!
function BGC.Player_Leaving_World(...)
	BGC.DebugPrint("Player has left PVP Battlegrounds")
	BGC.frame:UnregisterEvent("CHAT_MSG_ADDON");
	BGC.frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	BGC.frame:UnregisterEvent("PLAYER_LEAVING_WORLD");
	BGC.frame:UnregisterEvent("GROUP_ROSTER_UPDATE");
	BGC.frame:UnregisterEvent("UPDATE_BATTLEFIELD_SCORE");
	BGC.frame:UnregisterEvent("WORLD_MAP_UPDATE");
	
	-- TODO: Save off talent data to the saved variables
	BGC.DebugPrint("NOT clearing PD database!")
	--BGC.InitPD()
end


function BGC.BGMapID()
	local _, _, _, _, _, _, _, mapID = GetInstanceInfo()
	return mapID
	
end

function BGC.CheckNodeCache(mapID)
	if not BG_Control_SV.NodeCache[mapID] then
		BG_Control_SV.NodeCache[mapID] = {}
		local map = BG_Control_SV.NodeCache[mapID]
		BGC.InitNodeCache(map,mapID)
		BGC.CalculateNodeCacheRadius(map.node)
	end
end

function BGC.InitNodeCache(table,mapID)
	local index = 1
	table.node = {}
	for i=1,GetNumMapLandmarks() do
		local name, description, _, x, y, _, _, graveyardID = GetMapLandmarkInfo(i)
		-- Should we include graveyards?
		if 0 == graveyardID then
			if not table.node[index] then table.node[index] = {} end
			table.node[index].name = name
			table.node[index].description = description
			table.node[index].position = { x , y }
			table.node[index].landmarkID = i
			index = index + 1
		end
	
	end
end

function BGC.CalculateNodeCacheRadius(map)
	local radius = BGC.MAX_RADIUS
	local mapSize = BGC.TableSize(map)
	for i=1, mapSize do
		for j=i,mapSize do
			radius = BGC.Distance(map[i].position,(map[j].position))
		end
	end
	map.nodeRadius = radius
end

function BGC.ClosestNodeToPlayer(player)
	-- TODO: Distance is NOT in yards!  Do we dynamically figure it out to account for new BGs?
	local closestNode,range
	local nodes = BG_Control_SV.NodeCache[BGC.BGMapID()]
	local spellRange = 0
	if player.range then spellRange = player.range end

	local mapSize = BGC.TableSize(nodes.node)

	for i=1, mapSize do
		local tmpRange = BGC.Distance(player.position,nodes.node[i].position)
		if not range or tmpRange < range then
			closestNode = nodes.node[i].landmarkID
			range = tmpRange
		end
	end
	return closestNode, range + spellRange <= BGC.nodeRadius ,range
end

function BGC.TableSize(table)
	local count = 0
	for _ in pairs(table) do count = count + 1 end
	return count
end

function BGC.Distance(from,to)
	-- TODO: this is NOT in yards...
	if not from or not to then
		BGC.DebugPrint("Distance failure: From("..tostring(from)..") To("..tostring(to)..")")
		return nil
	end
	
	local x = from[1] - to[1]
	local y = from[2] - to[2]
	return math.sqrt(x * x + y * y)
end

function BGC.InitPD()
	PD={}
	PD.raidID = {}
	PD.damageMap = {}
	PD.healingMap = {}
	PD.NodeCount = {}
end


function BGC.Update_Battlefield_Score(...)
	--TODO: Use this to try to "prioritize" healers who are owning
end

function BGC.Combat_Log_Event_Unfiltered(...)
	local timeStamp, event, sourceName, sourceGUID, _, sourceFlags, destName, destGUID, _, destFlags, _, spellID = ...
	if BGC.IsPlayer(sourceFlags) and BGC.IsPlayer(destFlags) then
	
	
		if "SWING_DAMAGE" == event or "RANGE_DAMAGE" == event or "SPELL_DAMAGE" == event or "SPELL_HEAL" == event then
		if sourceGUID and not BGC.PlayerNameFromGUID(sourceGUID,sourceName,event) then
			BGC.DebugPrint("Combat Log Fun: "..sourceName.." | "..sourceGUID)
		end
		
		if destGUID and not BGC.PlayerNameFromGUID(destGUID,destName,event) then
			BGC.DebugPrint("Combat Log Fun: "..destName.." | "..destGUID)
		end
		end
	end
	BGC.ParseMessage(nil,event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID)
end

function BGC.ParseMessage(...)
	local isForign, event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags = ...
	-- TODO: Is it better to check flags BEFORE string compare? I assume so...
	if BGC.IsPlayer(sourceFlags) and BGC.IsPlayer(destFlags) then
		
		if "SWING_DAMAGE" == event then
			BGC.SwingDamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags, isForign)
		elseif "RANGE_DAMAGE" == event then
			local spellID= select(8,...)
			BGC.SpellDamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "SPELL_DAMAGE" == event then
			local spellID= select(8,...)
			BGC.SpellDamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "SPELL_HEAL" == event then
			local spellID= select(8,...)
			BGC.HealCast(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		end
	end
end

function BGC.Group_Roster_Update(...)
	BGC.DebugPrint("Modifying group size...")
	for i=1,40 do
		local raidID = "raid"..i;
		local name, realm = UnitName(raidID)
		if not realm then realm = GetRealmName() end
		PD.raidID[BGC.PlayerName(name,realm)] = raidID
	end
end

function BGC.SwingDamageDone(event,time,from,fromFlags,to,toFlags,isForign,fN,tN)
	local fromName, toName

	if BGC.IsHostile(fromFlags) then
		fromName = BGC.PlayerNameFromGUID(from)
		toName = BGC.PlayerNameFromGUID(to)
	elseif BGC.IsHostile(toFlags) then
		fromName = BGC.PlayerNameFromGUID(to)
		toName = BGC.PlayerNameFromGUID(from)
	end

	if not PD.damageMap[fromName] then PD.damageMap[fromName] = {} end
	local map = PD.damageMap[fromName]
	-- This is an update (used when doing Addon cross-talk)
	if not map.time or map.time < time then	
		local x, y = UnitPosition(BGC.RaidID(toName))
		map.time = time
		map.range = 0
		map.target = toName
		map.position = { x , y }
		BGC.UpdateDamage(map)
	end

	if not isForign then
		BGC.SendMessage(event,time,from,fromFlags,to,toFlags)
	end
end

function BGC.SendMessage(...)
	-- TODO: Make this better... have a function for damage, and a function for "other"
	--local msg = table.concat(table.pack(...),BGC.ADDON_MSG_SEPERATOR)
	--SendAddonMessage(BGC.ADDON_PREFIX, msg "BATTLEGROUND")
end

function BGC.Chat_Msg_Addon(...)
	local prefix, message = ...
	-- I could check the channel, but frankly I don't care... unless someone
	--	is trying to screw with this data, it's going to be fine
	if prefix == BGC.ADDON_PREFIX then
		BGC.ParseMessage(true,table.unpack(split(message,BGC.ADDON_MSG_SEPERATOR)))
	end
end

function BGC.SpellDamageDone(event,time,fromGUID,fromFlags,toGUID,toFlags,spellID, isForign)	
	local fromName, toName

	if BGC.IsHostile(fromFlags) then
		fromName = BGC.PlayerNameFromGUID(fromGUID)
		toName = BGC.PlayerNameFromGUID(toGUID)
	elseif BGC.IsHostile(toFlags) then
		fromName = BGC.PlayerNameFromGUID(toGUID)
		toName = BGC.PlayerNameFromGUID(fromGUID)
	end

	if not PD.damageMap[fromName] then PD.damageMap[fromName] = {} end
	local map = PD.damageMap[fromName]
	
	-- This is an update (used when doing Addon cross-talk)
	if not map.time or map.time < time then
		map.time = time
		map.range = BGC.SpellRange(spellID);
		map.target = toName
		if not BGC.RaidID(toName) then
			BGC.DebugPrint("Trying to unitPosition with name: "..toName.." and guid "..toGUID)
		end
		local x, y = UnitPosition(BGC.RaidID(toName))
		map.position = { x , y }
		BGC.UpdateDamage(map)
	end

	if not isForign then
		BGC.SendMessage(event,time,from,fromFlags,to,toFlags,spellID)
	end
end

function BGC.HealCast(event, time,from,fromFlags,to,toFlags,spellID, isForign)
	-- We don't care about actions for our players... only enemy
	--	Also, self-healing gives us NO data... ignore it
	if BGC.IsHostile(fromFlags) and from ~= to then
		local fromName = BGC.PlayerNameFromGUID(from)
		if not PD.healingMap[fromName] then PD.healingMap[fromName] = {} end
		local map = PD.healingMap[fromName]
		
		-- This is an update (used when doing Addon cross-talk)
		if not map.time or map.time < time then
			
			local toName = BGC.PlayerNameFromGUID(to)
			map.time = time
			map.range = BGC.SpellRange(spellID);
			map.target=toName;
			BGC.UpdateHealer(map)
		end

		-- If it's not from me... someone else got it too, so don't forward it
		if not isForign then
			BGC.SendMessage(event, time,from,fromFlags,to,toFlags,spellID)
		end
	end
end

function BGC.RaidID(name)
	if not PD.raidID[name] then
		--BGC.DebugPrint("Bad RaidID: "..name)
	end
	return PD.raidID[name]
end

function BGC.UpdateHealer(healingMap)
	local dm = PD.damageMap[healingMap.target]
	
	-- If we don't have a damage entry for this target... something's off
	if dm then
		local nodeID,isAtNode = BGC.ClosestNodeToPlayer(dm)
		BGC.UpdatePlayer(nodeID,isAtNode,healingMap)
	end
	
end

function BGC.UpdateDamage(damageMap)
	local nodeID,isAtNode = BGC.ClosestNodeToPlayer(damageMap)
	BGC.UpdatePlayer(nodeID,isAtNode,damageMap)
end

function BGC.UpdatePlayer(nodeID,isAtNode,player)
	if not PD.NodeCount[nodeID] then PD.NodeCount[nodeID] = {} end
	local map = PD.NodeCounts[nodeID]

	if player.lastNode then
		local nodeCounts = PD.NodeCount[player.lastNode.id]
		if player.lastNode.close then
			nodeCount.count = nodeCount.count - 1
		else
			nodeCount.nearby = nodeCount.nearby - 1
		end
	end 

	if player.node then
		player.lastNode = player.node
	end
 
	if isAtNode then
		map.count = map.count + 1
	else
		map.nearby = map.nearby + 1
	end

	local DEBUG= BG_Control_SV.NodeCache[BGC.GetMapID()][nodeID]
	BGC.DebugPrint(DEBUG.name .. ": ".. map.count .. " with " .. map.nearby .. " nearby.")

	-- TODO: Update the onhover/frame
end

-- TODO: Timer tick to "stale" old data

function BGC.PlayerName(name,realm)
	return name.."-"..realm
end

function BGC.SpellRange(id)
	local _, _, _, _, _, maxRange = GetSpellInfo(id)
	return maxRange
end

-- We need to do this, name doesn't include own realm and
--	guid is really long to compare... 
-- TODO: Benchmark GUID compare vs GetPlayerInfoByGUID()
-- TODO: Make a GUID Cache map in PD
function BGC.PlayerNameFromGUID(guid,name,event)
	local _,_,_,_,_,name,realm = GetPlayerInfoByGUID(guid)
	if not realm then realm = GetRealmName() end
	if guid and not name then 
		BGC.DebugPrint("GetPlayerInfoByGUID failure "..tostring(guid).." | "..tostring(name).." | "..tostring(event).." | "..realm)
	end
	return BGC.PlayerName(name,realm)
end

function BGC.IsHostile(flags)
	return bit.band(flags, COMBATLOG_OBJECT_REACTION_HOSTILE) > 0	
end

function BGC.IsPlayer(flags)	
	return bit.band(flags,COMBATLOG_OBJECT_TYPE_PLAYER) > 0
end

function BGC.Addon_Loaded(...)
	local name = ...
	if "BG_Control" == name then
		BGC.DebugPrint("Loaded BG_Control!")
	
	
		-- TODO: DELETE ME SOMETIME
		-- DEBUG LINEs
		BGC.DebugPrint("Deleting previous database for debugging")
		BG_Control_SV = {}
		BG_Control_SV.Options = {}
		BG_Control_SV.TalentCache = {}
		BG_Control_SV.MMRCache = {}
		BG_Control_SV.NodeCache = {}
	
		
		BGC.frame:UnregisterEvent("ADDON_LOADED");
		BGC.InitPD();
		-- This doesn't work for when you "leave"
		--hooksecurefunc("LeaveBattlefield",BGControlPublic.LeaveBattlefield);
	end
end

function BGC.OnEvent(self, event, ...)
	if "COMBAT_LOG_EVENT_UNFILTERED" == event then
		BGC.Combat_Log_Event_Unfiltered(...);
	elseif "CHAT_MSG_ADDON" == event then
		BGC.Chat_Msg_Addon(...);
	elseif "UPDATE_BATTLEFIELD_SCORE" == event then
		BGC.Update_Battlefield_Score(...);
	elseif "PLAYER_ENTERING_BATTLEGROUND" == event then
		BGC.Player_Entering_Battleground(...);
	elseif "GROUP_ROSTER_UPDATE" == event then
		BGC.Group_Roster_Update(...)
	elseif "WORLD_MAP_UPDATE" == event then
		BGC.World_Map_Update(...)
	elseif "PLAYER_LEAVING_WORLD" == event then
		BGC.Player_Leaving_World(...)
	elseif "ADDON_LOADED" == event then
		BGC.Addon_Loaded(...)
	end
end

function BGC.DebugPrint(str)
	-- This contains a curseforge macro to not package when deployed.
	--	See repository
	
end

-- Node Info is available only AFTER this event is fired
function BGC.World_Map_Update(...)
	BGC.DebugPrint("Updating World Map")
	BGC.CheckNodeCache(BGC.BGMapID())
	--BGC.Instance_Group_Size_Changed() -- This is in case of a reload
end


BGC.frame:SetScript("OnEvent",BGC.OnEvent);
BGC.frame:RegisterEvent("ADDON_LOADED");
BGC.frame:RegisterEvent("PLAYER_ENTERING_BATTLEGROUND");


